@@ -122,7 +122,6 @@ static int serialupdi_decode_sib(const PROGRAMMER *pgm, updi_sib_info *sib_info)
   memset(sib_info->nvm_string, 0, SIB_INFO_NVM_LENGTH+1);
   memset(sib_info->debug_string, 0, SIB_INFO_DEBUG_LENGTH+1);
   memset(sib_info->pdi_string, 0, SIB_INFO_PDI_LENGTH+1);
-  memset(sib_info->pdi_string, 0, SIB_INFO_PDI_LENGTH+1);
   memset(sib_info->extra_string, 0, SIB_INFO_EXTRA_LENGTH+1);
 
   memcpy(sib_info->family_string, sib_info->sib_string, SIB_INFO_FAMILY_LENGTH);
@@ -162,9 +161,19 @@ static int serialupdi_decode_sib(const PROGRAMMER *pgm, updi_sib_info *sib_info)
       updi_set_datalink_mode(pgm, UPDI_LINK_MODE_24BIT);
       break;
     case '3':
-      pmsg_notice("NVM type 3: 16-bit, page oriented\n");
+      pmsg_notice("NVM type 3: 24-bit, page oriented\n");
       updi_set_nvm_mode(pgm, UPDI_NVM_MODE_V3);
-      updi_set_datalink_mode(pgm, UPDI_LINK_MODE_16BIT);
+      updi_set_datalink_mode(pgm, UPDI_LINK_MODE_24BIT);
+      break;
+    case '4':
+      pmsg_notice("NVM type 4: 24-bit, word oriented\n");
+      updi_set_nvm_mode(pgm, UPDI_NVM_MODE_V4);
+      updi_set_datalink_mode(pgm, UPDI_LINK_MODE_24BIT);
+      break;
+    case '5':
+      pmsg_notice("NVM type 5: 24-bit, page oriented\n");
+      updi_set_nvm_mode(pgm, UPDI_NVM_MODE_V5);
+      updi_set_datalink_mode(pgm, UPDI_LINK_MODE_24BIT);
       break;
     default:
       pmsg_warning("unsupported NVM type: %c, please update software\n", sib_info->nvm_version);
@@ -275,6 +284,26 @@ static int serialupdi_wait_for_urow(const PROGRAMMER *pgm, unsigned int ms, urow
   return -1;
 }
 
+static int serialupdi_wait_for_nvmprog(const PROGRAMMER *pgm, unsigned int ms) {
+
+  unsigned long start_time;
+  unsigned long current_time;
+  uint8_t status;
+  start_time = avr_ustimestamp();
+  do {
+    if (updi_read_cs(pgm, UPDI_ASI_SYS_STATUS, &status) >= 0) {
+      if (status & (1 << UPDI_ASI_SYS_STATUS_NVMPROG)) {
+        return 0;
+      }
+    }
+    current_time = avr_ustimestamp();
+  } while ((current_time - start_time) < (ms * 1000));
+
+  pmsg_error("timeout waiting for device to enter NVMPROG mode\n");
+  return -1;
+}
+
+
 static int serialupdi_in_prog_mode(const PROGRAMMER *pgm, uint8_t *in_prog_mode) {
 /*
     def in_prog_mode(self):
@@ -355,6 +384,11 @@ def enter_progmode(self):
     return 0;
   }
 
+  if (serialupdi_reset(pgm, APPLY_RESET) < 0) {
+    pmsg_error("apply reset operation failed\n");
+    return -1;
+  }
+
   memcpy(buffer, UPDI_KEY_NVM, sizeof(buffer));
   if (updi_write_key(pgm, buffer, UPDI_KEY_64, sizeof(buffer)) < 0) {
     pmsg_error("writing NVM KEY failed\n");
@@ -368,8 +402,7 @@ def enter_progmode(self):
   pmsg_debug("key status: 0x%02X\n", key_status);
 
   if (!(key_status & (1 << UPDI_ASI_KEY_STATUS_NVMPROG))) {
-    pmsg_error("key was not accepted\n");
-    return -1;
+    pmsg_warning("key was not accepted\n");
   }
 
   if (serialupdi_reset(pgm, APPLY_RESET) < 0) {
@@ -387,12 +420,7 @@ def enter_progmode(self):
     return -1;
   }
 
-  if (serialupdi_in_prog_mode(pgm, &in_prog_mode) < 0) {
-    pmsg_error("checking UPDI NVM prog mode failed\n");
-    return -1;
-  }
-
-  if (!in_prog_mode) {
+  if (serialupdi_wait_for_nvmprog(pgm, 500) < 0) {
     pmsg_error("unable to enter NVM programming mode\n");
     return -1;
   }
@@ -514,9 +542,16 @@ static int serialupdi_write_userrow(const PROGRAMMER *pgm, const AVRPART *p, con
     return -1;
   }
 
-  if (updi_write_data(pgm, m->offset+addr, m->buf + addr, n_bytes) < 0) {
-    pmsg_error("writing USER ROW failed\n");
-    return -1;
+  if (n_bytes <= UPDI_MAX_REPEAT_SIZE) {
+    if (updi_write_data(pgm, m->offset+addr, m->buf + addr, n_bytes) < 0) {
+      pmsg_error("writing USER ROW failed\n");
+      return -1;
+    }
+  } else {
+    if (updi_write_data_words(pgm, m->offset+addr, m->buf + addr, n_bytes) < 0) {
+      pmsg_error("writing USER ROW failed\n");
+      return -1;
+    }
   }
 
   if (updi_write_cs(pgm, UPDI_ASI_SYS_CTRLA, (1 << UPDI_ASI_SYS_CTRLA_UROW_FINAL) |
@@ -639,8 +674,20 @@ static int serialupdi_initialize(const PROGRAMMER *pgm, const AVRPART *p) {
   }
 
   pmsg_notice("entering NVM programming mode\n");
-    /* try, but ignore failure */
-  serialupdi_enter_progmode(pgm);
+
+  /* try, but ignore failure */
+  /* It will always fail if the device is locked */
+  /* The device will be unlocked by erasing the chip after this. */
+  if (serialupdi_enter_progmode(pgm) == 0) {
+    /* If successful, you can run silicon check */
+    if (updi_read_data(pgm, p->syscfg_base+1, &value, 1) < 0) {
+      pmsg_error("Reading chip silicon revision failed\n");
+      return -1;
+    } else {
+      pmsg_debug("Received chip silicon revision 0x%02x\n", value);
+      pmsg_notice("Chip silicon revision: %x.%x\n", value >> 4, value & 0x0f);
+    }
+  }
 
   return 0;
 }
@@ -673,31 +720,68 @@ static int serialupdi_program_enable(const PROGRAMMER *pgm, const AVRPART *p) {
   return -1;
 }
 
+#define Return(...) do { pmsg_error(__VA_ARGS__); msg_error("\n"); return -1; } while (0)
+
 static int serialupdi_read_byte(const PROGRAMMER *pgm, const AVRPART *p, const AVRMEM *mem,
                                 unsigned long addr, unsigned char * value)
 {
+  pmsg_debug("%s(%s, 0x%04lx)\n", __func__, mem->desc, addr);
+  if(mem->size < 1)
+    Return("cannot read byte from %s %s owing to its size %d", p->desc, mem->desc, mem->size);
+
+  if(addr >= (unsigned long) mem->size)
+    Return("cannot read byte from %s %s as address 0x%04lx outside range [0, 0x%04x]",
+      p->desc, mem->desc, addr, mem->size-1);
+
+  if(mem_is_sib(mem)) {
+    if(addr >= SIB_INFO_STRING_LENGTH)
+      Return("cannot read byte from %s sib as address 0x%04lx outside range [0, 0x%04x]",
+        p->desc, addr, SIB_INFO_STRING_LENGTH-1);
+    if(!*updi_get_sib_info(pgm)->sib_string) // This should never happen
+      Return("cannot read byte from %s sib as memory not initialised", p->desc);
+    *value = updi_get_sib_info(pgm)->sib_string[addr];
+    return 0;
+  }
+
   return updi_read_byte(pgm, mem->offset + addr, value);
 }
 
 static int serialupdi_write_byte(const PROGRAMMER *pgm, const AVRPART *p, const AVRMEM *mem,
                                  unsigned long addr, unsigned char value)
 {
-  if (strstr(mem->desc, "fuse") != 0) {
+  pmsg_debug("%s(%s, 0x%04lx, 0x%02x)\n", __func__, mem->desc, addr, value);
+  if(mem->size < 1)
+    Return("cannot write byte to %s %s owing to its size %d", p->desc, mem->desc, mem->size);
+
+  if(addr >= (unsigned long) mem->size)
+    Return("cannot write byte to %s %s as address 0x%04lx outside range [0, 0x%04x]",
+      p->desc, mem->desc, addr, mem->size-1);
+
+  if (mem_is_a_fuse(mem) || mem_is_fuses(mem)) {
     return updi_nvm_write_fuse(pgm, p, mem->offset + addr, value);
   }
-  if (strcmp(mem->desc, "lock") == 0) {
+  if (mem_is_lock(mem)) {
     return updi_nvm_write_fuse(pgm, p, mem->offset + addr, value);
   }
-  if (strcmp(mem->desc, "eeprom") == 0) {
+  if (mem_is_eeprom(mem)) {
     unsigned char buffer[1];
     buffer[0]=value;
     return updi_nvm_write_eeprom(pgm, p, mem->offset + addr, buffer, 1);
   }
-  if (strcmp(mem->desc, "flash") == 0) {
+  if (mem_is_flash(mem)) {
     unsigned char buffer[1];
     buffer[0]=value;
     return updi_nvm_write_flash(pgm, p, mem->offset + addr, buffer, 1);
   }
+  // Read-only memories
+  if(mem_is_readonly(mem)) {
+    unsigned char is;
+    if(serialupdi_read_byte(pgm, p, mem, addr, &is) >= 0 && is == value)
+      return 0;
+
+    Return("cannot write to read-only memory %s of %s", mem->desc, p->desc);
+  }
+
   return updi_write_byte(pgm, mem->offset + addr, value);
 }
 
@@ -752,20 +836,20 @@ static int serialupdi_paged_write(const PROGRAMMER *pgm, const AVRPART *p, const
 
     while (remaining_bytes > 0) {
 
-      if (strcmp(m->desc, "eeprom")==0) {
+      if (mem_is_eeprom(m)) {
         rc = updi_nvm_write_eeprom(pgm, p, m->offset + write_offset, m->buf + write_offset, 
                                    remaining_bytes > m->page_size ? m->page_size : remaining_bytes);
-      } else if (strcmp(m->desc, "flash")==0) {
+      } else if (mem_is_flash(m)) {
         rc = updi_nvm_write_flash(pgm, p, m->offset + write_offset, m->buf + write_offset, 
                                   remaining_bytes > m->page_size ? m->page_size : remaining_bytes);
-      } else if (strcmp(m->desc, "userrow")==0) {
+      } else if (mem_is_userrow(m)) {
         rc = serialupdi_write_userrow(pgm, p, m, page_size, write_offset, 
                                       remaining_bytes > m->page_size ? m->page_size : remaining_bytes);
-      } else if (strcmp(m->desc, "fuses")==0) {
+      } else if (mem_is_fuses(m)) {
         pmsg_debug("page write operation requested for fuses, falling back to byte-level write\n");
         return -1;
       } else {
-        pmsg_error("invalid memory type: <%s:%d>, 0x%06X, %d (0x%04X)\n", m->desc, page_size, addr, n_bytes, n_bytes);
+        pmsg_error("invalid memory <%s:%d>, 0x%06X, %d (0x%04X)\n", m->desc, page_size, addr, n_bytes, n_bytes);
         rc = -1;
       }
 
@@ -780,17 +864,17 @@ static int serialupdi_paged_write(const PROGRAMMER *pgm, const AVRPART *p, const
     }
     return write_bytes;
   } else {
-    if (strcmp(m->desc, "eeprom")==0) {
+    if (mem_is_eeprom(m)) {
       rc = updi_nvm_write_eeprom(pgm, p, m->offset+addr, m->buf+addr, n_bytes);
-    } else if (strcmp(m->desc, "flash")==0) {
+    } else if (mem_is_flash(m)) {
       rc = updi_nvm_write_flash(pgm, p, m->offset+addr, m->buf+addr, n_bytes);
-    } else if (strcmp(m->desc, "userrow")==0) {
+    } else if (mem_is_userrow(m)) {
       rc = serialupdi_write_userrow(pgm, p, m, page_size, addr, n_bytes);
-    } else if (strcmp(m->desc, "fuses")==0) {
+    } else if (mem_is_fuses(m)) {
         pmsg_debug("page write operation requested for fuses, falling back to byte-level write\n");
         rc = -1;
     } else {
-      pmsg_error("invalid memory type: <%s:%d>, 0x%06X, %d (0x%04X)\n", m->desc, page_size, addr, n_bytes, n_bytes);
+      pmsg_error("invalid memory: <%s:%d>, 0x%06X, %d (0x%04X)\n", m->desc, page_size, addr, n_bytes, n_bytes);
       rc = -1;
     }
     return rc;
@@ -933,9 +1017,9 @@ static int serialupdi_parseextparms(const PROGRAMMER *pgm, const LISTID extparms
     extended_param = ldata(ln);
 
     if (sscanf(extended_param, "rtsdtr=%4s", rts_mode) == 1) {
-      if (strcasecmp(rts_mode, "low") == 0) {
+      if (str_caseeq(rts_mode, "low")) {
         updi_set_rts_mode(pgm, RTS_MODE_LOW);
-      } else if (strcasecmp(rts_mode, "high") == 0) {
+      } else if (str_caseeq(rts_mode, "high")) {
         updi_set_rts_mode(pgm, RTS_MODE_HIGH);
       } else {
         pmsg_error("RTS/DTR mode must be LOW or HIGH\n");
